/*
 * find.c
 * Copyright (C) LenaD 2010 <lenad89@list.ru>
 * 
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fnmatch.h>
#include <libgen.h>

void passgdir();
void mreaddir(DIR *dir);

char *name; /* pattern filename*/
char *currname; /*name catalog(file)*/

/* пока не используется
void *
xmalloc(size_t size)
{
	void *tmp = malloc(size);
	
	if (tmp == NULL) {		
		fprintf(stderr, "xmalloc: out of memory\n");
		exit(1);		
	}
	
	memset(tmp, 0, size);

	return tmp;	
}
*/
char *
xstrdup(const char *str)
{
        char *rs;
        size_t len;

        len = strlen(str);
        rs = malloc(len+1);

        if (rs == NULL) {
                fprintf(stderr, "xstrdup: out of memory\n");
                exit(1);
        }

        memcpy(rs, str, len);
        rs[len] = 0;

        return rs;
}

int is_correct(char *fname)
{
	if (name != NULL) {
		char *bname = basename(fname);	
		int is_fnm = fnmatch(name, bname, 0);
	
		if (is_fnm == FNM_NOMATCH) 
			return 0;
		else if (is_fnm == 0)
			return 1;
		else {
			fprintf(stderr, "is_correct: wrong pattern");
			exit(1);
		}	
	} else {
		return 1;
	}
}

void 
openderr(int direrr)
{
	switch (direrr){
	case EACCES:
	case ENOENT:
	case ENOTDIR:
		fprintf(stderr, "failed to open directory : "
	        	"%s :%s\n",currname,strerror(direrr));
		break;
	case ENOMEM:
	case ENFILE:
	case EMFILE:
		fprintf(stderr, "failed to open directory, fatal error : "
	        	"%s :%s\n",currname,strerror(direrr));
		exit(1);
		break;	
	}

}
void staterror(int sterr)
{
	if (sterr == ENOMEM) {
		fprintf(stderr, "failed to get stat : out of memory");
		exit(1);
	} else
		fprintf(stderr, "failed to get stat : "
	        	"%s :%s\n",currname,strerror(sterr));	 
}

void
passgdir()
{
	DIR* dir;
	struct stat pst;
	if (lstat(currname, &pst) == -1) {
		staterror(errno);
	} else {		
		if (is_correct(currname))	
			printf("%s\n",currname);	
		if (S_ISDIR(pst.st_mode)) {
			if ((dir = opendir(currname)) != NULL) {
				mreaddir(dir);
				closedir(dir);
			} else
				openderr(errno);		
		}
	}
}

void
get_fname(char * fname, size_t plen)
{
	char *tmp;
	size_t len;
	
	len = strlen(fname);
	
	if (currname[plen - 1] == '/')
		plen -= 1;
	
	tmp = realloc(currname, plen + len + 2);
	
	 if (tmp == NULL) {
                fprintf(stderr, "get_fname: out of memory\n");
                exit(1);
        }
	
	tmp[plen] = '/';
	memcpy(tmp + plen + 1 , fname, len);
	tmp[len + plen + 1] = '\0';
	currname = tmp;
}

void
mreaddir(DIR *dir)
{
	struct dirent *tmpd;
	size_t plen = strlen(currname);
	while ((tmpd = readdir(dir)) != NULL) {
		if (strcoll(tmpd->d_name, ".") && strcoll(tmpd->d_name, "..")) {
			get_fname(tmpd->d_name, plen);		
			passgdir();
		}		
				
	}
}

void
go_path(int cpath, char **argv)
{	
	int i;
	
	if (cpath > 0) {
		for(i = 0; i < cpath; i++) {	
			currname = xstrdup(argv[i+1]);
			passgdir();
			free(currname);			
		}
	} else {
		currname = xstrdup(".");
		passgdir();
		free(currname);			
	}
}

void get_opt(int numopt, int argc, char **argv)
{
	int check_name = 0;

	if (numopt < argc && !strcoll(argv[numopt], "-name")) {
		check_name = 1;
		if (++numopt < argc){	
			name = xstrdup(argv[numopt]);
			++numopt;	
		}
	}
	if ( (numopt < argc) || (check_name && name == NULL  )) {
		fprintf(stderr, "error input: find [path][-name:pattern]\n");
		exit(1);
	}	
}

int
main(int argc, char **argv)
{	
	int i;
	int cpath;
	
	for(i = 1, cpath = 0; i < argc && argv[i][0] != '-'; i++)
		cpath++;
	
	get_opt(i, argc, argv);
	go_path(cpath, argv);
	free(name);
	return 0;
}	
